/*
	This file is part of lgob.

	lgob is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	lgob is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with lgob.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2009 Lucas Hermann Negri
*/

#include "lglib.h"

static void special_type_new(lua_State* L, const gchar* mt, gpointer ptr)
{
	if(ptr)
	{
		Object* obj = lua_newuserdata(L, sizeof(Object));
		obj->pointer = ptr;
		obj->need_unref = TRUE;
		lua_getfield(L, LUA_REGISTRYINDEX, mt);
		lua_setmetatable(L, -2);
	}
	else
		lua_pushnil(L);
}

int lgob_timeout_add(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TNUMBER);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TFUNCTION);
	
	#ifdef IDEBUG
	g_debug("Adding a timeout!");
	#endif
		
	/* Ref the function and the user data */
	int ud_ref = -1;
	
	if(lua_gettop(L) >= 4)
	{
		ud_ref = luaL_ref(L, LUA_REGISTRYINDEX);
		lua_pushvalue(L, 3);
	}	
	
	int function_ref = luaL_ref(L, LUA_REGISTRYINDEX);	
	
	/* Create the callback data */
	Data* data = g_malloc(sizeof(Data));
	data->L = L;
	data->function_ref = function_ref;
	data->ud_ref = ud_ref;
	data->signal_info.param_types = NULL;
	data->signal_info.n_params = 0;
	data->signal_info.return_type = G_TYPE_BOOLEAN;
	
	/* Connect and send the id to Lua */
	guint res = priv_timeout_add(lua_tointeger(L, 1), lua_tointeger(L, 2), data);
	lua_pushnumber(L, res);
	
	return 1;
}

int lgob_timeout_remove(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TNUMBER);
	
	#ifdef IDEBUG
	g_debug("Removing a timeout!");
	#endif
	
	guint id = (guint)lua_tointeger(L, 1);
	g_source_remove(id);
	
	return 0; 
}

int lgob_timeout_idle_add(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TNUMBER);
	luaL_checktype(L, 2, LUA_TFUNCTION);
	
	#ifdef IDEBUG
	g_debug("Adding a idle timeout!");
	#endif
		
	/* Ref the function and the user data */
	int ud_ref = -1;
	
	if(lua_gettop(L) >= 3)
	{
		ud_ref = luaL_ref(L, LUA_REGISTRYINDEX);
		lua_pushvalue(L, 2);
	}	
	
	int function_ref = luaL_ref(L, LUA_REGISTRYINDEX);	
	
	/* Create the callback data */
	Data* data = g_malloc(sizeof(Data));
	data->L = L;
	data->function_ref = function_ref;
	data->ud_ref = ud_ref;
	data->signal_info.param_types = NULL;
	data->signal_info.n_params = 0;
	data->signal_info.return_type = G_TYPE_BOOLEAN;
	
	/* Connect and send the id to Lua */
	guint res = priv_idle_add(lua_tointeger(L, 1), data);
	lua_pushnumber(L, res);
	
	return 1;
}

int lgob_main_loop_new(lua_State* L)
{
	GMainLoop* loop = g_main_loop_new(NULL, FALSE);
	special_type_new(L, "glibMainLoopMT", loop);

	return 1;
}

int lgob_main_loop_run(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* ptr = lua_touserdata(L, 1);
	g_main_loop_run(ptr->pointer);

	return 0;
}

int lgob_main_loop_iterate(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* ptr = lua_touserdata(L, 1);
	int top = lua_gettop(L);
	gboolean block = TRUE;

	if(top > 1)
		block = lua_toboolean(L, 2);

	gboolean res = g_main_context_iteration(g_main_loop_get_context(ptr->pointer), block);
	lua_pushboolean(L, res);

	return 1;
}

int lgob_main_loop_pending(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* ptr = lua_touserdata(L, 1);

	lua_pushboolean(L, g_main_context_pending(g_main_loop_get_context(ptr->pointer)));
	return 1;
}

int lgob_main_loop_quit(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* ptr = lua_touserdata(L, 1);

	g_main_loop_quit(ptr->pointer);

	return 0;
}

int lgob_main_loop_free(lua_State* L)
{
	Object* ptr = lua_touserdata(L, 1);
	g_main_loop_unref(ptr->pointer);

	return 0;
}

#if GLIB_MINOR_VERSION >= 8

int lgob_environment_get_host_name(lua_State* L)
{
	lua_pushstring(L, g_get_host_name());
	return 1;	
}

#endif

#if GLIB_MINOR_VERSION >= 10

int lgob_date_set_time_t(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	g_date_set_time_t(obj->pointer, lua_tointeger(L, 2)); 
	
	return 0;
}

#endif

#if GLIB_MINOR_VERSION >= 12

int lgob_time_val_from_iso8601(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TSTRING);
	Object* obj = lua_touserdata(L, 1);
	
	gboolean res = g_time_val_from_iso8601(lua_tostring(L, 2), obj->pointer);
	lua_pushboolean(L, res);
	
	return 1;
}

int lgob_time_val_to_iso8601(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	gchar* res = g_time_val_to_iso8601(obj->pointer);
	lua_pushstring(L, res);
	g_free(res);
	
	return 1;
}

int lgob_base64_encode(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	
	/* Get the string and its size */
	size_t len;
	const guchar* str = (guchar*)lua_tolstring(L, 1, &len);
	
	/* Do the conversion */
	gchar* coded = g_base64_encode(str, len);
	
	/* Push and release */
	lua_pushstring(L, coded);
	g_free(coded);
	
	return 1;
}

int lgob_base64_decode(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	
	/* Decode */
	size_t len;
	guchar* decoded = g_base64_decode(lua_tostring(L, 1), &len); 
	
	/* Push and release */
	lua_pushlstring(L, (gchar*)decoded, len);
	g_free(decoded);
	
	return 1;
}

#endif

#if GLIB_MINOR_VERSION >= 16

int lgob_checksum_get_md5(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	gchar* sum = g_compute_checksum_for_string(G_CHECKSUM_MD5, lua_tostring(L, 1), lua_strlen(L, 1));
	lua_pushstring(L, sum);
	g_free(sum);
	
	return 1; 
}

int lgob_checksum_get_sha1(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	gchar* sum = g_compute_checksum_for_string(G_CHECKSUM_SHA1, lua_tostring(L, 1), lua_strlen(L, 1));
	lua_pushstring(L, sum);
	g_free(sum);
	
	return 1;	
}

int lgob_checksum_get_sha256(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	gchar* sum = g_compute_checksum_for_string(G_CHECKSUM_SHA256, lua_tostring(L, 1), lua_strlen(L, 1));
	lua_pushstring(L, sum);
	g_free(sum);
	
	return 1;
}

#endif

int lgob_environment_get_user_name(lua_State* L)
{
	lua_pushstring(L, g_get_user_name());
	return 1;	
}

int lgob_environment_get_real_name(lua_State* L)
{
	lua_pushstring(L, g_get_real_name());
	return 1;	
}

int lgob_environment_get_user_data_dir(lua_State* L)
{
	lua_pushstring(L, g_get_user_data_dir());
	return 1;	
}

int lgob_environment_get_user_cache_dir(lua_State* L)
{
	lua_pushstring(L, g_get_user_cache_dir());
	return 1;	
}

int lgob_environment_get_user_config_dir(lua_State* L)
{
	lua_pushstring(L, g_get_user_config_dir());
	return 1;	
}

int lgob_environment_get_tmp_dir(lua_State* L)
{
	lua_pushstring(L, g_get_tmp_dir());
	return 1;	
}

int lgob_environment_get_current_dir(lua_State* L)
{
	gchar* curr = g_get_current_dir();
	lua_pushstring(L, curr);
	g_free(curr);
	
	return 1;	
}

int lgob_environment_get_home_dir(lua_State* L)
{
	lua_pushstring(L, g_get_home_dir());
	return 1;	
}

int lgob_path_is_absolute(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	lua_pushboolean(L, g_path_is_absolute(lua_tostring(L, 1)));
	
	return 1;
}

int lgob_path_skip_root(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	lua_pushstring(L, g_path_skip_root(lua_tostring(L, 1)));
	
	return 1;
}

int lgob_path_get_basename(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	gchar* path = g_path_get_basename(lua_tostring(L, 1));
	lua_pushstring(L, path);
	g_free(path);
	
	return 1;
}

int lgob_path_get_dirname(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	gchar* path = g_path_get_dirname(lua_tostring(L, 1));
	lua_pushstring(L, path);
	g_free(path);
	
	return 1;
}

int lgob_path_find_program(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	gchar* path = g_find_program_in_path(lua_tostring(L, 1));
	lua_pushstring(L, path);
	g_free(path);
	
	return 1;
}

int lgob_date_new(lua_State* L)
{
	GDate* ptr = g_date_new();
	special_type_new(L, "glibDateMT", ptr);
	
	return 1;
}

int lgob_date_free(lua_State* L)
{
	Object* obj = lua_touserdata(L, 1);
	g_date_free(obj->pointer);
	
	return 0;
}

int lgob_date_set_dmy(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	g_date_set_dmy(obj->pointer, lua_tointeger(L, 2), lua_tointeger(L, 3), 
		lua_tointeger(L, 4));
	
	return 0;
}

int lgob_date_set_julian(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	g_date_set_julian(obj->pointer, lua_tointeger(L, 2));
	
	return 0;
}

int lgob_date_set_parse(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	g_date_set_parse(obj->pointer, lua_tostring(L, 2));
	
	return 0;
}

int lgob_date_is_leap_year(lua_State* L)
{
	lua_pushboolean(L, g_date_is_leap_year(lua_tointeger(L, 1)));
	return 1;
}

int lgob_date_compare(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	
	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);

	lua_pushinteger(L, g_date_compare(obj1->pointer, obj2->pointer));
	return 1;
}

int lgob_date_valid(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	lua_pushboolean(L, g_date_valid(obj->pointer));
	return 1;
}

int lgob_date_get_day(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	lua_pushinteger(L, g_date_get_day(obj->pointer));
	return 1;
}

int lgob_date_get_month(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	lua_pushinteger(L, g_date_get_month(obj->pointer));
	return 1;
}

int lgob_date_get_year(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	lua_pushinteger(L, g_date_get_year(obj->pointer));
	return 1;
}

int lgob_rand_free(lua_State* L)
{
	Object* obj = lua_touserdata(L, 1);
	g_rand_free(obj->pointer);
	
	return 0;
}

int lgob_rand_new(lua_State* L)
{
	GRand* ptr = g_rand_new();
	special_type_new(L, "glibRandMT", ptr);
	
	return 1;
}

int lgob_rand_set_seed(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	Object* obj = lua_touserdata(L, 1);
	
	g_rand_set_seed(obj->pointer, lua_tointeger(L, 2));
	
	return 0;
}

int lgob_rand_boolean(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	gboolean res = g_rand_boolean(obj->pointer);
	lua_pushboolean(L, res);
	
	return 1;
}

int lgob_rand_int(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	int top = lua_gettop(L);
	guint res = -1;
	
	if(top == 1)
		res = g_rand_int(obj->pointer);
	else
	{
		luaL_checktype(L, 2, LUA_TNUMBER);
		luaL_checktype(L, 3, LUA_TNUMBER);
		res = g_rand_int_range(obj->pointer, lua_tointeger(L, 2), lua_tointeger(L, 3));
	}
	
	lua_pushnumber(L, res);
	return 1;
}

int lgob_rand_double(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	int top = lua_gettop(L);
	gdouble res = -1;
	
	if(top == 1)
		res = g_rand_double(obj->pointer);
	else
	{
		luaL_checktype(L, 2, LUA_TNUMBER);
		luaL_checktype(L, 3, LUA_TNUMBER);
		res = g_rand_double_range(obj->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3));
	}
	
	lua_pushnumber(L, res);
	return 1;
}

int lgob_time_val_free(lua_State* L)
{
	Object* obj = lua_touserdata(L, 1);
	g_free(obj->pointer);
	
	return 0;
}

int lgob_get_current_time(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	g_get_current_time(obj->pointer);
	return 0;
}

int lgob_time_val_new(lua_State* L)
{
	GTimeVal* ptr = g_malloc(sizeof(GTimeVal));
	special_type_new(L, "glibTimeValMT", ptr);
	
	return 1;
}

int lgob_time_val_add(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	Object* obj = lua_touserdata(L, 1);
	
	g_time_val_add(obj->pointer, lua_tointeger(L, 2));	
	return 0;
}

int lgob_timer_free(lua_State* L)
{
	Object* obj = lua_touserdata(L, 1);	
	g_timer_destroy(obj->pointer);
	
	return 1;
}

int lgob_timer_new(lua_State* L)
{
	GTimer* ptr = g_timer_new();
	special_type_new(L, "glibTimerMT", ptr);
	
	return 1;
}

int lgob_timer_start(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	g_timer_start(obj->pointer);
	
	return 0;
}

int lgob_timer_stop(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	g_timer_stop(obj->pointer);
	
	return 0;
}

int lgob_timer_continue(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	g_timer_continue(obj->pointer);
	
	return 0;
}

int lgob_timer_elapsed(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	gulong micro;
	gdouble sec = g_timer_elapsed(obj->pointer, &micro);
	
	lua_pushnumber(L, sec);
	lua_pushnumber(L, micro);
	
	return 2;
}

int lgob_utf8_validate(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	size_t len;
	const char* str = lua_tolstring(L, 1, &len);
	
	gboolean res = g_utf8_validate(str, len, NULL);
	lua_pushboolean(L, res);
	
	return 1;
}

int lgob_utf8_strup(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	size_t len;
	const char* str = lua_tolstring(L, 1, &len);
	
	gchar* res = g_utf8_strup(str, len);
	lua_pushstring(L, res);
	g_free(res);
	
	return 1;
}

int lgob_utf8_strdown(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	size_t len;
	const char* str = lua_tolstring(L, 1, &len);
	
	gchar* res = g_utf8_strdown(str, len);
	lua_pushstring(L, res);
	g_free(res);
	
	return 1;
}

int lgob_utf8_casefold(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	size_t len;
	const char* str = lua_tolstring(L, 1, &len);
	
	gchar* res = g_utf8_casefold(str, len);
	lua_pushstring(L, res);
	g_free(res);
	
	return 1;
}


int lgob_utf8_normalize(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	luaL_checktype(L, 2, LUA_TNUMBER);
	size_t len;
	const char* str = lua_tolstring(L, 1, &len);
	
	gchar* res = g_utf8_normalize(str, len, (GNormalizeMode)lua_tointeger(L, 2));
	lua_pushstring(L, res);
	g_free(res);
	
	return 1;
}

int lgob_utf8_strlen(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	size_t len;
	const char* str = lua_tolstring(L, 1, &len);
	
	glong res = g_utf8_strlen(str, len);
	lua_pushinteger(L, res);
	
	return 1;
}

int lgob_utf8_collate(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	luaL_checktype(L, 2, LUA_TSTRING);
	
	gint res = g_utf8_collate(lua_tostring(L, 1), lua_tostring(L, 2));
	lua_pushinteger(L, res);
	
	return 1;
}

int lgob_utf8_collate_key(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	size_t len;
	const char* str = lua_tolstring(L, 1, &len);
	
	gchar* res = g_utf8_collate_key(str, len);
	lua_pushstring(L, res);
	g_free(res);
	
	return 1;
}

int lgob_utf8_from_filename(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	size_t len;
	const char* str = lua_tolstring(L, 1, &len);
	
	gchar* res = g_filename_to_utf8(str, len, NULL, NULL, NULL);
	lua_pushstring(L, res);
	g_free(res);
	
	return 1;
}

int lgob_utf8_to_filename(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	size_t len;
	const char* str = lua_tolstring(L, 1, &len);
	
	gchar* res = g_filename_from_utf8(str, len, NULL, NULL, NULL);
	lua_pushstring(L, res);
	g_free(res);
	
	return 1;
}

int lgob_convert(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	luaL_checktype(L, 2, LUA_TSTRING);
	luaL_checktype(L, 3, LUA_TSTRING);
	
	size_t len, bread, bwrite;
	const char* str = lua_tolstring(L, 1, &len);
	GError* errors = NULL;
	
	char* res = g_convert(str, len, lua_tostring(L, 2), lua_tostring(L, 3), 
		&bread, &bwrite, &errors);
		
	if(errors)
	{
		lua_pushboolean(L, FALSE);
		lua_pushinteger(L, errors->code);
		lua_pushstring(L, errors->message);
		g_error_free(errors);
		if(res) g_free(res);
		return 3;
	}
	
	lua_pushlstring(L, res, bwrite);
	g_free(res);
	
	return 1;
}

int lgob_get_charset(lua_State* L)
{
	const char* charset;
	gboolean res = g_get_charset(&charset);
	
	lua_pushboolean(L, res);
	lua_pushstring(L, charset);
	return 2;
}

int lgob_thread_init(lua_State* L)
{
	g_thread_init(NULL);
	return 0;
}
